; URIDISD.ASM      URIDIUM Sound Driver      J.DAVE ROGERS. Liverpool. 1987.


;**************************************************************************
SOUNDCOD        EQU       40000     ;machine code load address. length 758
SOUNDATA        EQU       40760     ;datablock load address.   length 1430
;**************************************************************************


                NOLIST

PITCHTAB        EQU       SOUNDATA        ;200 bytes
ENVSINDEX       EQU       PITCHTAB+200    ;70  bytes
TUNESINDEX      EQU       ENVSINDEX+70    ;104 bytes
ENVS            EQU       TUNESINDEX+110  ;304 bytes
TUNEBLOCS       EQU       ENVS+307        ;

envcounts       EQU 0    ;offsets within channel stores:
entcounts       EQU 2
envptr          EQU 4
entptr          EQU 6
env0            EQU 8
ent0            EQU 10
envadr          EQU 12
entadr          EQU 14

chanactive      EQU 16   ;1
maincount       EQU 17   ;1
dataptr         EQU 18   ;2
conptr          EQU 20   ;2
ptrrefr         EQU 22   ;2
pitchshift      EQU 24   ;1
noiseflag       EQU 25   ;1
noisedis        EQU 26   ;1
ch_enable       EQU 27   ;1
advolreg        EQU 28   ;2
stophang        EQU 30   ;1

                ORG  SOUNDCOD
                WRITE "URIDISD.BIN"
start           DEFL $

;------------------------------------------------------------------------------

JUMPBLOCK       JP PLAY          ;used for convenience in Basic Demo
                JP SILENCE
POKE_BASIC      LD E,1           ;tunebloc
                LD A,1           ;chan
                JR STARTACHAN

;------------------------
;  "E" MUST CONTAIN THE NUMBER OF THE TUNEBLOC TO BE STARTED
;  CALL CH1, PUTS TUNE ON CHANNEL 1, CH2 = CHANNEL 2
;  CALL CHANALT, DYNAMICALLY ALLOCATES TO CHANS 2/3 ALTERNATELY

CH1             LD A,1
                JR STARTACHAN
CH2             LD A,2
                JR STARTACHAN
CHANALT         LD A,2
                XOR 1
                LD (CHANALT+1),A   ;A GOES 2,3,2,3....

STARTACHAN      LD IX,STORE1      ;E contains tunebloc num, A contains chan num
                DEC A
                JR Z ,IXfound
                LD IX,STORE2
                DEC A
                JR Z,IXfound
                LD IX,STORE3
IXfound         CALL FINDBLOC           ;RETURNS BLOCAD IN HL
                LD   (IX+DATAPTR),L     ;set pointers for this channel
                LD   (IX+DATAPTR+1),H
                LD   (IX+CONPTR),L
                LD   (IX+CONPTR+1),H
                LD   (IX+PTRREFR),L     ;ptr refresh
                LD   (IX+PTRREFR+1),H
                LD (IX+CHANACTIVE),1
                LD (IX+MAINCOUNT),1
                LD A,0
                LD (IX+PITCHSHIFT),A
                LD (CHORUSFLAG),A
                RET
;-----------------------
SILENCE         XOR A
                LD  (STORE1+CHANACTIVE),A    ;channel-active-flags 0
                LD  (STORE2+CHANACTIVE),A
                LD  (STORE3+CHANACTIVE),A
                LD (BUFF+8),A                ;volume registers 0
                LD (BUFF+9),A
                LD (BUFF+10),A
                LD   A,1+2+4+8+16+32         ;all gates shut
                LD   (BUFF+7),A              ;follow through to allregist

ALLREGIST       XOR  A
                LD   HL,BUFF+13              ;register buffer
                LD   DE,#F40D                ;D=Psg data port, E=counter
nextreg         LD   B,D
                OUT  (C),E                   ;output register No
                LD   BC,#F6*256+%11000000
                OUT  (C),C                   ;tell PSG to latch reg No
                OUT  (C),A                   ;PSG to inactive
                LD   B,D
                LD   C,(HL)                  ;C=register data
                DEC  HL
                OUT  (C),C                   ;output register data
                LD   BC,#F6*256+%10000000
                OUT  (C),C                   ;tell PSG to read reg data
                OUT  (C),A                   ;PSG to inactive
                DEC  E
                JP   P,nextreg
                RET
;------------
FINDBLOC        LD   A,E
                ADD  A,A
                ADD  A,TUNESINDEX AND #00FF
                LD   L,A
                ADC  A,TUNESINDEX AND #FF00 /256
                SUB  L
                LD   H,A              ;HL now points to 2byte value in table
                LD   E,(HL)
                INC  HL
                LD   D,(HL)           ;DE=value(offset)
                LD HL,TUNEBLOCS
                ADD HL,DE             ;HL=address of new tunebloc
                RET

FINDENV         LD   A,E
                ADD  A,A
                ADD  A,ENVSINDEX AND #00FF       ;16 bit add using 8 bit ops
                LD   L,A
                ADC  A,ENVSINDEX AND #FF00 / 256
                SUB  L
                LD   H,A              ;HL points to 2byte offset
                LD   E,(HL)
                INC  HL
                LD   D,(HL)           ;DE=offset
                LD HL,ENVS
                ADD HL,DE             ;HL=envelope address
                RET

;--------------------MAIN CALL--------EVERY 1/50 SECOND-----------------------

PLAY            CALL ALLREGIST     ;throw all buffer into sound chip registers
                LD   IX,STORE1     ;---do channel 1 (IX points to chan1 store)
                LD   IY,BUFF       ;  (IY holds addr of pitchstore for chan1)
                CALL TESTCHAN
                LD   IX,STORE2     ;---do channel 2
                LD   IY,BUFF+2
                CALL TESTCHAN
                LD   IX,STORE3     ;---do channel 3
                LD   IY,BUFF+4
                CALL TESTCHAN
                CALL DOENSE        ;do noise envelope
                LD A,(CHORUSFLAG)  ;test if autochorus on
                OR A
                RET Z
                LD HL,(BUFF)
                LD BC,4
                ADD HL,BC
                LD (BUFF+4),HL
                RET              ;-------------->>>>>>> out

;FOR EACH OF THE 3 CHANNELS:

TESTCHAN        LD  A,(IX+CHANACTIVE)   ;CHAN ACTIVE?
                OR A
                RET Z                   ;do nothing if channel not active
                DEC  (IX+maincount)     ;duration countdown, new note when 0
                JP NZ,DOENV_ENT
                LD (IX+STOPHANG),30

READDECIDE      LD  H,(IX+dataptr+1)    ;get data ptr address
                LD  L,(IX+dataptr)
DECI            LD   A,(HL)             ;A=peek(ptr)
                INC  HL
                LD   E,(HL)             ;E=peek(ptr+1)
                INC HL                  ;inc and store ptr for next time
CONTPTR         LD (IX+DATAPTR+1),H
                LD (IX+DATAPTR),L
                DEC (IX+STOPHANG)
                RET Z     ;gets out of possible endless loops after 30 passes
;-------------------------------------------------------------
; THE FOLLOWING ROUTINES RESPOND TO DATA AND JUMP BACK TO READDECIDE OR DECI,
; EVENTUALLY EXITING VIA "NEWNOTE" OR "KILLCHAN" ROUTINES

SUBSELECT       CP   0             ;A=VALUE
                JP   Z,DOPAUSE     ;pitch 0 = pause
                CP   101
                JP   C,NEWNOTE     ;LESS THAN 100=normal note
                CP   220
                JP   C,SINGLENOTE
                CP   228
                JR   Z,NOISE       ;add noise to a note
                CP   225
                JR   Z,KILLCHAN    ;shut chan after e.g.short sound effects
                CP   233
                JR   Z,NEWENT      ;set new tone-envelope for current channel
                CP   232
                JR   Z,NEWENV      ;set new volume-env for current chan
                CP   234
                JR   Z,NEWENSE     ;set new noise-envelope
                CP   229
                JR   Z,GONEWBLOCK  ;play a given tunebloc,then return to 1st
                CP   230
                JR   Z,NEWKEY      ;change musical key for current chan
                CP   235
                JR   Z,CHORUSONOFF ;switch autochorus on or off
                CP   255
                JR   Z,RETCONBLOC  ;end-of-tunebloc marker
                CP   252
                RET  C        ;SAFETY NET
                              ;vals 252,253,254=follow through:

STARTLOOP       PUSH IX            ;252=STARTCHAN1, 253=CHAN2, 254=CHAN3
                SUB 251            ;A=CHAN E=TUNEBLOC
                CALL STARTACHAN
                POP IX
                JR READDECIDE

NOISE           LD   A,E
                LD   (BUFF+6),A        ;noisepitch
                LD   (IX+noiseflag),1
                JR   DECI

NEWKEY          LD   (IX+pitchshift),E
                JR   DECI

NEWENT          CALL FINDENV           ;returns with envelope address in HL
                LD (IX+ENTADR),L
                LD (IX+ENTADR+1),H
                JR READDECIDE

NEWENV          CALL FINDENV
                LD (IX+ENVADR),L
                LD (IX+ENVADR+1),H
                JR READDECIDE

NEWENSE         CALL FINDENV
                LD (ENSSTORE+12),HL
                JP READDECIDE

GONEWBLOCK      LD   (IX+conptr+1),H
                LD   (IX+conptr),L    ;leaving bloc so store ptr
                CALL FINDBLOC
                JP   DECI

RETCONBLOC      LD   H,(IX+conptr+1)  ;(ptr re-enters original "conductor" bloc)
                LD   L,(IX+conptr)
                LD   A,(HL)
                INC A                 ;TEST FOR 255,end-of-conductor-bloc marker
                JR   NZ,NOTEND
                LD   H,(IX+ptrrefr+1)        ;refresh
                LD   L,(IX+ptrrefr)
notend          JP   DECI

KILLCHAN        LD (IX+CHANACTIVE),0
killvol         LD H,(IX+ADVOLREG+1)
                LD L,(IX+ADVOLREG)     ;HL=address of vol register,current chan
                LD (HL),0
                RET

CHORUSONOFF     LD HL,CHORUSFLAG
                LD (HL),E
                JP READDECIDE

SINGLENOTE      AND 127
                DEC HL
                LD E,10
                JP CONTPTR

NEWNOTE         ADD  A,(IX+pitchshift)       ;A=notenum
DOPAUSE         LD   (IX+maincount),E        ;E=duration

                ADD  A,A                     ;get new pitch from pitchtable
                ADD  A,PITCHTAB AND #00FF
                LD   E,A
                ADC  A,PITCHTAB AND #FF00 /256
                SUB  E
                LD   D,A
                LD   A,(DE)
                LD   (IY+0),A        ;IY=address of pitchstore for current chan
                INC  DE
                LD   A,(DE)
                LD   (IY+1),A

                call killvol ;zero volume of current chan

                PUSH IX              ;RESTART ENV/ent
                POP  DE              ;DE=start of chan store
                LD   HL,8
                ADD  HL,DE           ;HL=start of chan store+8
                LDI
                LDI
                LDI
                LDI
                LDI
                LDI
                LDI
                LDI

                DEC  (IX+noiseflag)      ;test noiseflag
                LD  (IX+NOISEFLAG),0     ;RESET EVERY TIME!
                LD  L,(IX+NOISEDIS)      ;DEFAULT DISABLE
                JR   NZ,GATING

NOISEon         LD HL,(ENSSTORE+envadr)  ;restart noise envelope
                LD (ENSSTORE+envptr),HL
                LD HL,00
                LD (ENSSTORE),HL         ;L=0,no noisedisable

GATING          LD   A,(BUFF+7)
                AND  (IX+ch_enable)      ;chan ENable (pitch AND noise on)
                OR   L                   ;noise DISable
                AND 32+16+8+4+2+1        ;SAFETY, top 2 bits must remain off
                LD   (BUFF+7),A          ;FINAL RESULT IN A!
;------------------------------------------------------
                           ;IX points to volume-env workspace for current chan
DOENV_ENT       CALL ENGENERAL         ;returns with "volume change+128" in C
                LD   H,(IX+advolreg+1)
                LD   L,(IX+advolreg)
                LD   A,(HL)
                ADD  A,C
                SUB  128
                LD   (HL),A
DOENT           INC  IX  ;make IX work on tone-envelope workspace,2 bytes along
                INC  IX
                CALL ENGENERAL         ;returns with "pitch change+128" in C
                LD   H,(IY+1)          ;get pitch of note
                LD   L,(IY+0)
                LD A,H
                OR L
                RET Z         ;no ent if pitch zero(pause)
                ADD  HL,BC
                LD   C,128             ;deduct 128, allows negative changes
                SBC  HL,BC
                LD   (IY+1),H          ;replace in current channel pitch store
                LD   (IY+0),L
                RET

DOENSE          LD   IX,ENSSTORE       ;IX now = noise envelope workspace
                CALL ENGENERAL         ;returns with "noise shift+128" in C
                LD   A,(BUFF+6)        ;get current noise value
                ADD  A,C
                SUB  128               ;add shift, minus 128
                LD   (BUFF+6),A        ;replace
                CP 17
                RET C
noiseoff        LD A,(BUFF+7)
                OR 32+16+8
                LD (BUFF+7),A
                RET

ENGENERAL       PUSH IX       ;subroutine to do envelope and return value
                POP  HL                ;points to counts
                LD   D,(IX+ENVPTR+1)   ;points to envelope
                LD   E,(IX+ENVPTR)
                INC  (HL)              ;inc count
                LD   A,(DE)
                SUB  (HL)              ;compare to envelope
                LD   BC,128             ;DEFAULT (C= 128 = 0 )
                RET  NZ                ;count not reached
                LD   (HL),A            ;A=0, resets count
                INC DE                 ;move along envelope
                LD  A,(DE)
                LD  C,A                ;C HOLDS INCREMENT
                INC  DE                ;point at repeats number
                INC  HL
                INC  (HL)              ;point at repeats count
                LD   A,(DE)
                SUB  (HL)              ;compare
                RET  NZ                ;count not reached
                LD   (HL),A            ;A=0, resets count;
                INC  DE                ;move to next section of envelope
                LD   A,(DE)
                INC  A                 ;test for A=255 (end of envelope marker)
                JR   NZ,nextsect
                LD   D,(IX+envadr+1)   ;refresh ptr to start of envelope
                LD   E,(IX+envadr)
nextsect        LD   (IX+envptr+1),D
                LD   (IX+envptr),E
                RET
;--------------------------------------------------------

BUFF      DW  00       ;a pitch     ;holds replica of sound chip registers
          DW  00       ;b pitch
          DW  00       ;c pitch
          DB  0        ;noise pitch
          DB  32+16+8  ;gates
          DB  0        ;a vol
          DB  0        ;b vol
          DB  0        ;c vol
          DW  100       ;hard env len
          DB  10        ;hard env type

CHORUSFLAG  DB  0

ensstore  DW 00 ;ENSCOUNTS    ;noise envelope workspace
          DW 10               ;dummy spacer to give same format as chan stores
          DW 00 ;ENSPTR
          DW 04
          DW 00 ;ENS0
          DW 18
          DW 00 ;ENSADR
;----------------------------------
;channel stores

STORE1    DW  00        ;env counts
          DW  00        ;ENT COUNTS
          DW  00        ;ENV PTR
          DW  00        ;ENT PTR

          DW  00        ;ENV ZEROERS
          DW  00        ;ENT ZEROERS
          DW  00        ;ENV ADDR
          DW  00        ;ENT ADDR

          DB  0         ;CHAN ACTIVE
          DB  0         ;MAIN COUNT
          DW  00        ;MAIN PTR
          DW  00        ;CON PTR
          DW  00        ;PTR REFRESH
          DB  0         ;PITCHSHIFT
          DB  0         ;NOISE FLAG
          DB   8        ;NOISE DIS
          DB  32+16+4+2 ;chan ENable
          DW  BUFF+8    ;AD OF VOL REG
          DB  0

STORE2  DW  00
        DW  00
        DW  00
        DW  00

        DW  00
        DW  00
        DW  00
        DW  00

        DB  0
        DB  0
        DW  00
        DW  00
        DW  00
        DB  0
        DB  0
        DB  16        ;NOISE DIS
        DB  32+8+4+1  ;chan ENable
        DW  BUFF+9    ;AD OF VOL REg
        DB  0

STORE3  DW  00
        DW  00
        DW  00
        DW  00

        DW  00
        DW  00
        DW  00
        DW  00

        DB  0
        DB  0
        DW  00
        DW  00
        DW  00
        DB  0
        DB  0
        DB  32       ;NOISE DIS
        DB  16+8+2+1 ;CHAN ENABLE
        DW  BUFF+10  ;AD OF VOL REG
        DB  0
;------------------------------------------------
FINISH          DEFL $
                LIST
                CLOSE
START           DEFL START
FINISH          DEFL FINISH
LENGTH          DEFL FINISH-START


;*********************************